!=========================================================================
! Copyright (C) GemTalk Systems 1986-2015.  All Rights Reserved.
!
! $Id: rcindexbucket.gs,v 1.10 2008-01-09 19:02:35 stever Exp $
!
! Superclass Hierarchy:
!   RcIndexBucket, RcCollisionBucket, CollisionBucket, Array,
!   SequenceableCollection, Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods RcIndexBucket
removeallclassmethods RcIndexBucket
protectmethods

category: 'For Documentation Installation only'
classmethod: RcIndexBucket
installDocumentation

< unprotected >

| doc txt |
doc := GsClassDocumentation _newForPrivateGsClass: self.

txt := (GsDocText new) details:
'An RcIndexBucket is an RcCollisionBucket that is used by the indexing
 subsystem to store a collection of key/value/pathTerm triples.  It provides
 reduced-conflict support for indexes.' .
doc documentClassWith: txt.

self description: doc.
%

!==========================================================================
! Below this point: only method definitions in fileout order
!==========================================================================
! ------------------- Class methods for RcIndexBucket
category: 'Instance Creation'
classmethod: RcIndexBucket
new

"Returns a new instance of RcIndexBucket."
<protected>

^ self new: 0

%
! ------------------- Instance methods for RcIndexBucket
category: 'Adding'
method: RcIndexBucket
add: newObject

"Protected version of Array | add:."

<protected primitive: 203>
self _primitiveFailed: #add: .
self _uncontinuableError

%
category: 'Adding'
method: RcIndexBucket
addLast: newObject

"Protected version of Array | addLast:."

<protected primitive: 203>
self _primitiveFailed: #addLast: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
at: aKey put: aValue term: pathTerm root: indexDict logging: aBoolean

"Places the given key/value/pathTerm triplet in the receiver.  Returns whether
 the key/value pair already exists in the receiver."
<protected>

| index valSet value offset offsetMinus1 offsetMinus2 |
index := self _binarySearchCoveringKey: aKey term: pathTerm.
offset := self valueOffsetFor: index.

" if it goes on the end ... "
index > numElements
  ifTrue: [
    " grow the receiver if necessary "
    self _basicSize < offset
      ifTrue: [
        " do not want bucket to grow greater than a page "
        self _basicSize: ((offset + 15) min: 2002).
        " initialize the cache if necessary "
        numElements == 0
          ifTrue: [
            1 to: self _numCacheSlots do: [ :i |
              self _basicAt: i put: 0
            ]
          ]
      ].

    self _basicAtOffset: offset putKey: aKey term: pathTerm value: aValue.
    numElements := numElements + 1.
    ^ false
  ].

" if key is already in bucket "
offsetMinus1 := offset - 1 .
offsetMinus2 := offset - 2 .
( aKey == (self _rcAt: offsetMinus2 ) _and:
[ pathTerm == (self _rcAt: offsetMinus1) ] )

  ifTrue: [
    value := self _rcAt: offset.
    " see if there are already multiple values "
    (self _isBucketValueBagAt: offset)
      ifTrue: [
        " put the BucketValueBag in the reduced-conflict read set "
        System _addToRcReadSet: value includingAllNodes: true.
        " put an entry in the conflict objects dictionary "
        aBoolean
          ifTrue: [ System redoLog addConflictObject: value for: indexDict ].

        value add: aValue
      ]
      ifFalse: [ " create value Bag "
        valSet := BucketValueBag basicNew.
        valSet assignToSegment: self segment.
        valSet add: value.
        valSet add: aValue.
        self _basicAt: offset put: valSet.

        self insertedBucketValueBagAt: offset.
      ].
    ^ true
  ]
  ifFalse: [ " key not found "
    " move entries to the right (hard coded for performance) "
    self _insertAt: offsetMinus2
      from: nil "insert nils" fromStart: 1 fromEnd: 3 
      numToMoveDown: (numElements * 3) + self _numCacheSlots - offsetMinus2 + 1 .

    self _basicAtOffset: offset putKey: aKey term: pathTerm value: aValue.
    self insertedEntryAt: offset.

    " only increment numElements if a key is being added "
    numElements := numElements + 1.
    ^ false
  ]

%
category: 'Updating'
method: RcIndexBucket
at: anIndex putKey: aKey

"Stores the key aKey into the key part of the key/value/pathTerm triplet
 referenced by atIndex."
<protected>

self _basicAt: (anIndex * "self entrySize" 3) - 2 put: aKey

%
category: 'Updating'
method: RcIndexBucket
at: anIndex putTerm: pathTerm

"Stores the pathTerm into the pathTerm part of the key/value/pathTerm triplet
 referenced by atIndex."
<protected>

self _basicAt: (anIndex * "self entrySize" 3 - 1) put: pathTerm

%
category: 'Updating'
method: RcIndexBucket
at: anIndex putValue: aValue

"Stores the value aValue into the value part of the key/value/pathTerm triplet
 referenced by atIndex."
<protected>

self _basicAt: (anIndex * "self entrySize" 3) put: aValue

%
category: 'Accessing'
method: RcIndexBucket
at: aKey term: pathTerm

"Returns the value that corresponds aKey/pathTerm by searching the elements in
 the bucket.  if no such key/value/pathTerm triplet exists, raise an error."
<protected>

| index |
index := self searchForKey: aKey term: pathTerm.
index ~~ nil
    ifTrue: [ ^ self valueAt: index ]
    ifFalse: [ ^ self _errorKeyNotFound: aKey ]

%
category: 'Accessing'
method: RcIndexBucket
at: aKey term: pathTerm ifAbsent: aBlock

"Returns the value that corresponds to aKey/pathTerm by searching the elements
 in the bucket.  If no such key/value/pathTerm triplet exists, returns the
 result of evaluating the zero-argument block aBlock."
<protected>

| index |
index := self searchForKey: aKey term: pathTerm.
^ index ~~ nil
    ifTrue: [ self valueAt: index ]
    ifFalse: [ aBlock value ]
%
category: 'Accessing'
method: RcIndexBucket
at: aKey term: pathTerm otherwise: aValue

"Returns the value that corresponds aKey/pathTerm by searching the elements in
 the bucket.  If no such key/value/pathTerm triplet exists, returns aValue."

<protected primitive: 180>
self _primitiveFailed: #at:term:otherwise: .
self _uncontinuableError

%
category: 'Accessing'
method: RcIndexBucket
basicAt: anIndex

"Protected version of Object | basicAt:."
<protected>

^self _basicAt: anIndex

%
category: 'Updating'
method: RcIndexBucket
basicAt: anIndex put: aValue

"Protected version of Object | basicAt:put:."
<protected>

^self _basicAt: anIndex put: aValue

%
category: 'Copying'
method: RcIndexBucket
copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex

"Protected version of Array | copyFrom:to:into:startingAt:."

<protected primitive: 808"with rcvrClass">

^ super copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex

%
category: 'Constants'
method: RcIndexBucket
entrySize

"Returns three, the size of an entry in the bucket (key, term, value)."
<unprotected>

^ 3

%
category: 'Accessing'
method: RcIndexBucket
firstAt: aKey term: pathTerm otherwise: aValue

"Returns the value that corresponds aKey/pathTerm by searching the elements in
 the bucket.  If no such key/value/pathTerm triplet exists, returns aValue.
 If the value is a BucketValueBag, returns the first object in the Bag."
<protected>

| val |
val := self at: aKey term: pathTerm otherwise: aValue.
(BucketValueBag _hasInstance: val)
  ifTrue: [ ^ val _at: 1 ]
  ifFalse: [ ^ val ]

%
category: 'Copying'
method: RcIndexBucket
insertAll: aSeqColl at: anIndex

"Protected version of Array | insertAll:at:."

<protected primitive: 806"with rcvrClass">
^ super insertAll: aSeqColl at: anIndex

%
category: 'Updating'
method: RcIndexBucket
insertedBucketValueBagAt: offset

"Does nothing."

<unprotected>

%
category: 'Updating'
method: RcIndexBucket
insertedEntryAt: offset

"Does nothing."

<unprotected>

%
category: 'Accessing'
method: RcIndexBucket
instVarAt: anIndex

"Protected version of Object | instVarAt:."

<protected primitive: 811"with rcvrClass">

| numInstVars |
(anIndex _isInteger)
ifTrue:
[
   numInstVars:= self class allInstVarNames size.
   (numInstVars == 0)
   ifTrue:
      [ ^ self _error: #rtErrNoInstVars].
   ((anIndex >= 1) & (anIndex <= numInstVars))
   ifFalse: "out of bounds"
      [ ^ self _errorIndexOutOfRange: anIndex]
]
ifFalse: "anIndex not an integer"
   [ ^ self _errorNonIntegerIndex: anIndex
].

self _primitiveFailed: #instVarAt: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
instVarAt: anIndex put: aValue

"Protected version of Object | instVarAt:put:."

<protected primitive: 804"with rcvrClass">
(self class allInstVarNames size == 0)
  ifTrue: [ ^ self _error: #rtErrNoInstVars].
(anIndex _isInteger)
  ifTrue: [((anIndex < 1) | (anIndex > self class allInstVarNames size))
        ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]] "out of bounds"
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #instVarAt:put: .
self _uncontinuableError

%
category: 'Accessing'
method: RcIndexBucket
keyAt: index

"Returns the key at the specified index."
<protected>

^ self _rcAt: ("self entrySize" 3 * index - 2)

%
category: 'Updating'
method: RcIndexBucket
removedBucketValueBagAt: offset

"Does nothing."

<unprotected>

%
category: 'Updating'
method: RcIndexBucket
removedEntryAt: offset

"Does nothing."

<unprotected>

%
category: 'Copying'
method: RcIndexBucket
removeFrom: startIndex to: stopIndex

"Protected version of Array | removeFrom:to:."

<protected primitive: 801>

^ super removeFrom: startIndex to: stopIndex

%
category: 'Removing'
method: RcIndexBucket
removeKey: aKey
value: aValue
term: pathTerm
root: indexDict
logging: aBoolean
ifAbsent: aBlock

"Removes the key/value/pathTerm triplet with the given key and path term.
 Executes the zero-argument block if not found.  Returns whether there are any
 more values for the key/term."
<protected>

| keyIndex valOrSet eSize delOffset offset |

keyIndex := self searchForKey: aKey term: pathTerm.

keyIndex == nil
  ifTrue: [ " key not found "
    aBlock value.
    ^ false
  ]
  ifFalse: [
    valOrSet := self valueAt: keyIndex.
    " do not put in rcReadSet (to fix bug 10796)
    System _addToRcReadSet: valOrSet includingAllNodes: true.
    "

    offset := self valueOffsetFor: keyIndex.
    (self _isBucketValueBagAt: offset)
      ifTrue: [ " remove the value from the bucket value set "

        aBoolean
          ifTrue: [
            " put an entry in the conflict objects dictionary "
            System redoLog addConflictObject: valOrSet for: indexDict
          ].

        valOrSet remove: aValue ifAbsent: aBlock.
        " if only 1 left, move it to the receiver "
        valOrSet size == 1
          ifTrue: [
            " do not put in rcReadSet
            System _addToRcReadSet: (valOrSet _at: 1) includingAllNodes: true.
            "
            self at: keyIndex putValue: (valOrSet _at: 1).

            self removedBucketValueBagAt: offset.
          ].
        ^ true
      ]
      ifFalse: [ " key has a single value "
        (aValue == valOrSet)
          ifTrue: [
            eSize := "self entrySize"  3 .
            delOffset := keyIndex - 1 * eSize + self _numCacheSlots + 1 .
            self _deleteNoShrinkFrom: delOffset to: delOffset + eSize - 1 .
            self removedEntryAt: offset.
            numElements := numElements - 1.
            ^ false
          ]
          ifFalse: [ " values do not match "
            aBlock value.
            ^ true
          ]
      ]
  ]

%
category: 'Updating'
method: RcIndexBucket
reset

"Resets all entries to nil and the number of elements to zero."

<protected>
super size: self _numCacheSlots.
numElements := 0.
1 to: self _numCacheSlots do: [ :i |
  self _basicAt: i put: 0
].
System _addToRcReadSet: self includingAllNodes: false.

%
category: 'Searching'
method: RcIndexBucket
searchForKey: aKey term: pathTerm

"Returns the index that matches the given key/pathTerm.  If not found, returns
 nil."

<protected primitive: 175>
self _primitiveFailed: #searchForKey:term: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
size: anInteger

"Protected version of Object | size:."

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError

%
category: 'Accessing'
method: RcIndexBucket
tableSize

"Returns the number of key/value/pathTerm pairs my current size has the
 capacity for."
<protected>

^ ((self _basicSize - self _numCacheSlots) // 3) max: 0

%
category: 'Accessing'
method: RcIndexBucket
termAt: index

"Returns the term at the specified index."
<protected>

^ self _rcAt: ("self entrySize" 3 * index - 1)

%
category: 'Accessing'
method: RcIndexBucket
valueAt: index

"Returns the value at the specified index."
<protected>

^ self _rcAt: ("self entrySize" 3 * index)

%
category: 'Accessing'
method: RcIndexBucket
valueOffsetFor: index

"Returns the value at the specified index."
<protected>

^ "self entrySize" 3 * index

%
category: 'Adding'
method: RcIndexBucket
_addAllFromNsc: aBag

"Protected version of Array | _addAllFromNsc:."

<protected primitive: 202>
self _primitiveFailed: #_addAllFromNsc: .
self _uncontinuableError

%
category: 'Accessing'
method: RcIndexBucket
_at: anIndex

"Protected version of Object | _at:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_at: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
_at: anIndex put: aValue

"Protected version of Object | _at:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_at:put: .
self _uncontinuableError

%
category: 'Accessing'
method: RcIndexBucket
_basicAt: anIndex

"Protected version of Object | _basicAt:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_basicAt: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
_basicAt: anIndex put: aValue

"Protected version of Object | _basicAt:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_basicAt:put: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
_basicAtOffset: offset putKey: aKey term: pathTerm value: aValue
"<offset> is the actual offset into the bucket, of the value.  Insert the 
key/term/value in the slots appropriately offset from the <offset>.
This method is hard coded for performance "
<protected>

self _basicAt: offset - 2 put: aKey.
self _basicAt: offset - 1 put: pathTerm.
self _basicAt: offset put: aValue.

%
category: 'Updating'
method: RcIndexBucket
_basicSize: anInteger

"Protected version of Object | _basicSize:."

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError

%
category: 'Searching'
method: RcIndexBucket
_binarySearchCoveringKey: aKey term: pathTerm

"Returns the index for the first entry in which aKey is found utilizing a
 binary search.  This is the first entry whose key >= aKey."

<protected primitive: 115>
self _primitiveFailed: #_binarySearchCoveringKey:term: .
self _uncontinuableError

%
category: 'Private'
method: RcIndexBucket
_canonicalizeSymbolAt: offset oldSymbol: old newSymbol: newSymbol

"Override the method, but do nothing here because other updates will fix it."

<protected>

^self

%
category: 'Removing'
method: RcIndexBucket
_deleteNoShrinkFrom: startIndex to: stopIndex

"Protected version of Array | _deleteNoShrinkFrom:to:."

<protected primitive: 802 "with rcvrClass">
^ super _deleteNoShrinkFrom: startIndex to: stopIndex

%
category: 'Copying'
method: RcIndexBucket
_insertAt: destOffset
from: anArray
fromStart: startOffset
fromEnd: endOffset
numToMoveDown: numToMove

"Protected version of Array | #_insertAt:from:fromStart:fromEnd:numToMoveDown:."

<protected primitive: 805"with rcvrClass">
^ super _insertAt: destOffset from: anArray fromStart: startOffset
            fromEnd: endOffset numToMoveDown: numToMove

%
category: 'Testing'
method: RcIndexBucket
_isBucketValueBagAt: offset

"Returns whether the object at the given offset in the receiver is a
BucketValueBag."

<protected>
^ BucketValueBag _hasInstance: (self _rcAt: offset)

%
category: 'Constants'
method: RcIndexBucket
_numCacheSlots

"Returns the number of initial slots reserved for cached values."
<unprotected>

^ 0

%
category: 'Accessing'
method: RcIndexBucket
_primitiveAt: anIndex

"Protected version of Object | _primitiveAt:."

<protected primitive: 89>

(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_primitiveAt: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
_primitiveAt: anIndex put: aValue

"Protected version of Object | _primitiveAt:put:."

<protected primitive: 281>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
((anIndex < 1) | (anIndex > (self _primitiveSize + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_primitiveAt:put: .
self _uncontinuableError

%
category: 'Reduced Conflict Support'
method: RcIndexBucket
_rcAt: anIndex

"Protected version of Object | _rcAt:."

<protected primitive: 92>
self _primitiveFailed: #_rcAt: .
self _uncontinuableError

%
category: 'Updating'
method: RcIndexBucket
_refreshBucketCacheReportingInto: aString

"Does nothing."

<unprotected>

%
category: 'Searching'
method: RcIndexBucket
_stBinarySearchCoveringKey: aKey term: pathTerm

"Returns the index for the first entry in which aKey is found utilizing a
 binary search.  This is the first entry whose key >= aKey."
<protected>

| lowPt midPt highPt keyAlias pathTermAlias |

numElements == 0 ifTrue: [ ^ 1 ].

keyAlias := aKey identityHash.
pathTermAlias := pathTerm identityHash.
lowPt := self _numCacheSlots + 1.
highPt := numElements.
[ lowPt <= highPt ] whileTrue: [
    midPt := (lowPt + highPt) quo: 2.

    keyAlias == (self keyAt: midPt) identityHash
        ifTrue: [
            pathTermAlias <= (self termAt: midPt) identityHash
                ifTrue: [ highPt := midPt - 1 ]
                ifFalse: [ lowPt := midPt + 1 ]
        ]
        ifFalse: [
            keyAlias < (self keyAt: midPt) identityHash
                ifTrue: [ highPt := midPt - 1 ]
                ifFalse: [ lowPt := midPt + 1 ]
        ]
].

( keyAlias > (self keyAt: midPt) identityHash _or:
[ keyAlias == (self keyAt: midPt) identityHash _and:
[ pathTermAlias > (self termAt: midPt) identityHash ] ] )
            ifTrue: [  ^ midPt + 1 ].

^ midPt

%


unprotectmethods

